unit BTCpuCount;

 {* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
  * BTCpuCount (CPU Counting Util-Unit, incl. Hyper-Threading-Technology)   *
  *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*
  * Copyright (c) 2002 - 2010 by Martin Offenwanger (www.dsplayer.de)  	    *
  *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*
  * Mozilla Public License Version 1.1:                                     *
  *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*
  * The contents of this file are used with permission, subject to the      *
  * Mozilla Public License Version 1.1 (the "License"); you may             *
  * not use this file except in compliance with the License. You may        *
  * obtain a copy of the License at                                         *
  * http://www.mozilla.org/MPL/MPL-1.1.html                                 *
  *                                                                         *
  * Software distributed under the License is distributed on an             *
  * "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or          *
  * implied. See the License for the specific language governing            *
  * rights and limitations under the License.                               *
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * }
{
@author(Martin Offenwanger: mo@danfort.dk, coder@dsplayer.de)
@created(Nov 24, 2010)
@lastmod(Nov 20, 2010)
}

interface

// *****************************************************************************
//
//  function BTGetCPUCount(out f_LogicalNum: Byte;
//                                    out f_PhysicalNum: Byte): TBTHTStatusFlag;
//
// Note: 1) f_LogicalNum = Number of logical processors per PHYSICAL PROCESSOR.
//       If you want to count the total number of logical processors,
//       multiply this number with the total number of physical
//       processors (f_PhysicalNum)
//
//       2) To detect whether hyper-threading is enabled or not is to see how
//       many logical ID exist per single physical ID in APIC
//
//       3) For systems that don't support hyper-threading like AMD or PIII
//       and below. the variable LogicalNum will be set to 1 (which means
//       number of logical processors equals to number of physical processors.)
//
//       4) Hyper-threading cannot be detected when application cannot access
//       all processors in the system. The number of physical processors will
//       be set to 255.  Make sure to enable ALL physical processors at startup
//       of windows, and applications calling this function, CPUCount, are NOT
//       restricted to run on any particular logical or physical processors
//       (can run on ALL processors.)
//
//       5) Windows currently can handle up to 32 processors.
// *****************************************************************************

uses Windows, SysUtils;

type
  // status flag type definition
  TBTHTStatusFlag = Byte;
const
  // possible status flag states (TBTHTStatusFlag)
  HT_NOT_CAPABLE = 0;
  HT_ENABLED = 1;
  HT_DISABLED = 2;
  HT_SUPPORTED_NOT_ENABLED = 3;
  HT_CANNOT_DETECT = 4;

function BTGetCPUCount(out f_LogicalNum: Byte;
  out f_PhysicalNum: Byte): TBTHTStatusFlag;

implementation

const
  HT_BIT = $10000000;                (* EDX[28]  Bit 28 is set if HT is
                                        supported *)
  FAMILY_ID = $0F00;                 (* EAX[11:8] Bit 8-11 contains family
                                        processor ID. *)
  PENTIUM4_ID = $0F00;
  EXT_FAMILY_ID = $0F00000;          (* EAX[23:20] Bit 20-23 contains extended
                                        family processor ID *)
  NUM_LOGICAL_BITS = $00FF0000;      (* EBX[23:16] Bit 16-23 in ebx contains
                                        the number of logical processors per
                                        physical processor when execute cpuid
                                        with eax set to 1 *)

  INITIAL_APIC_ID_BITS = $FF000000;  (* EBX[31:24] Bits 24-31 (8 bits) return
                                        the 8-bit unique initial APIC ID for
                                        the processor this code is running on.
                                        Default value = 0xff if HT is not
                                        supported *)


function BTIsHyperThreadingSupported: Cardinal;
  function VendorIDxToString(f_VendorIDx: Cardinal): string;
  var
    l_Count: Integer;
    lp_VendorIdx: PByte;
  begin
    SetLength(Result, 4);
    lp_VendorIdx := @f_VendorIDx;
    for l_Count := 1 to 4 do begin
      Result[l_Count] := Char(lp_VendorIdx^);
      inc(lp_VendorIdx, 1);
    end;
  end;
var
  l_Count: Integer;
  l_Regedx, l_Regeax: Cardinal;
  l_VendorID0, l_VendorID1,l_VendorID2: Cardinal;
begin
  // init fields
  l_Regedx := 0;
  l_Regeax := 0;
  l_VendorID0 := 0;
  l_VendorID1 := 0;
  l_VendorID2 := 0;

  try // Verify cpuid instruction is supported
    asm
      push      eax                     // Save affected registers
      push      ebx
      push      edx
      push      ecx
      xor       eax, eax                // call cpuid with eax = 0
      cpuid                             // Get vendor id string
      mov       l_VendorID0, ebx
      mov       l_VendorID1, edx
      mov       l_VendorID2, ecx
      mov       eax, 1                  // call cpuid with eax = 1
      cpuid
      mov       l_Regeax, eax           // eax contains family processor type
      mov       l_Regedx, edx           (* edx has info about the availability
                                           of hyper-Threading *)
      pop       ecx                     // Restore registers
      pop       edx
      pop       ebx
      pop       eax
    end;
  except
    Result := HT_NOT_CAPABLE;  // cpuid is unavailable
    exit;
  end;

  // check for Genuine Intel with hyper-Threading technology
  if ((l_Regeax and FAMILY_ID) = PENTIUM4_ID) or
    (Cardinal(l_Regeax and EXT_FAMILY_ID) >= 1) then
      if VendorIDxToString(l_VendorID0) = 'Genu' then
        if VendorIDxToString(l_VendorID1) = 'ineI' then
          if VendorIDxToString(l_VendorID2) = 'ntel' then begin
            // Genuine Intel with hyper-Threading technology
            Result := l_Regedx and HT_BIT;
            exit;
          end;

  Result := HT_NOT_CAPABLE; // Not genuine Intel processor
end;

function BTGetLogicalProcPerPhysicalProc: Byte;
var
  l_Regebx: Cardinal;
begin
   l_Regebx := 0;
   if (not (Cardinal(BTIsHyperThreadingSupported) >= 1)) then begin
     Result := 1;
     exit;
   end;
   asm
     push       eax
     push       ebx
     mov        eax, 1
     cpuid
     mov        l_Regebx, ebx
     pop        ebx
     pop        eax
   end;
   Result := Byte((l_Regebx and NUM_LOGICAL_BITS) shr 16);
end;

function BTGetAPIC_ID: Byte;
var
  l_Regebx: Cardinal;
begin
   l_Regebx := 0;
   if (not (Cardinal(BTIsHyperThreadingSupported) >= 1)) then begin
     // set Result to 255 (Hint: Byte(-1) = 255)
     Result := Byte(-1);  // HT not supported, Logical processor = 1
     exit;
   end;
   asm
     push       eax
     push       ebx
     mov        eax, 1
     cpuid
     mov        l_Regebx, ebx
     pop        ebx
     pop        eax
   end;
   Result := Byte((l_Regebx and INITIAL_APIC_ID_BITS) shr 24)
end;

function BTGetCPUCount(out f_LogicalNum: Byte; out f_PhysicalNum: Byte): Byte;
var
  l_StatusFlag: Byte;
  l_SysInfo: SYSTEM_INFO;
  l_HT_Enabled: Byte;
  lh_CurrentProcessHandle: THandle;
  l_dwProcessAffinity: DWORD;
  l_dwSystemAffinity: DWORD;
  l_dwAffinityMask: DWORD;
  l_i, PHY_ID_MASK, PHY_ID_SHIFT: Byte;
  APIC_ID, LOG_ID, PHY_ID: Byte;
  BITWISE_NEOGATION_OF_PHY_ID_MASK: Byte;
begin
  l_HT_Enabled := 0;
  l_StatusFlag := 0;
  l_SysInfo.dwNumberOfProcessors := 0;
  GetSystemInfo(l_SysInfo);

  // Number of physical processors in a non-Intel system
  // or in a 32-bit Intel system with Hyper-Threading technology disabled
  f_PhysicalNum := Byte(l_SysInfo.dwNumberOfProcessors);

  if Cardinal(BTIsHyperThreadingSupported) >= 1 then begin
     f_LogicalNum := BTGetLogicalProcPerPhysicalProc;
     if (f_LogicalNum >= 1) then begin (* >1 Doesn't mean HT is enabled
                                          in the BIOS *)
       (* Calculate the appropriate  shifts and mask based on the
          number of logical processors. *)
       l_i := 1;
       PHY_ID_MASK := $FF;
       PHY_ID_SHIFT := 0;
       while (l_i < f_LogicalNum) do begin
         l_i := l_i * 2;
         PHY_ID_MASK := PHY_ID_MASK shl 1;
         inc(PHY_ID_SHIFT);
       end;

       lh_CurrentProcessHandle := GetCurrentProcess();
       GetProcessAffinityMask(lh_CurrentProcessHandle, l_dwProcessAffinity,
         l_dwSystemAffinity);

       (* Check if available process affinity mask is equal to the
          available system affinity mask *)
        if l_dwProcessAffinity <> l_dwSystemAffinity then begin
          l_StatusFlag := HT_CANNOT_DETECT;
          f_PhysicalNum := Byte(-1); // set to 255
          Result := l_StatusFlag;
          exit;
        end;

        l_dwAffinityMask := 1;
        while ((l_dwAffinityMask <> 0) and
                (l_dwAffinityMask <= l_dwProcessAffinity)) do begin
          // Check if this CPU is available
          if Cardinal(l_dwAffinityMask and l_dwProcessAffinity) >= 1 then begin
            if SetProcessAffinityMask(lh_CurrentProcessHandle,
                        l_dwAffinityMask) then begin
              Sleep(0); // Give OS time to switch CPU
              APIC_ID := BTGetAPIC_ID;
              // bitwieses complement errechnen
              BITWISE_NEOGATION_OF_PHY_ID_MASK := not PHY_ID_MASK;
              LOG_ID  := APIC_ID and BITWISE_NEOGATION_OF_PHY_ID_MASK;
              PHY_ID := APIC_ID shr PHY_ID_SHIFT;
              if (LOG_ID <> 0) then
                l_HT_Enabled := 1;
            end;
          end;
          l_dwAffinityMask := l_dwAffinityMask shl 1;
        end;

        // Reset the processor affinity
        SetProcessAffinityMask(lh_CurrentProcessHandle, l_dwProcessAffinity);
        if f_LogicalNum = 1 then begin
          l_StatusFlag := HT_DISABLED;
        end else begin
          if l_HT_Enabled >= 1 then begin
            // Total physical processors in a Hyper-Threading enabled system.
            f_PhysicalNum := Trunc(f_PhysicalNum / f_LogicalNum);
	    l_StatusFlag := HT_ENABLED;
          end else begin
            l_StatusFlag := HT_SUPPORTED_NOT_ENABLED;
          end;
        end;
     end;
  end else begin
    // Processors do not have Hyper-Threading technology
    l_StatusFlag := HT_NOT_CAPABLE;
    f_LogicalNum := 1;
  end;
  Result := l_StatusFlag;
end;

end.
